-------------------------------------------------------------------------------
--! @file       CipherCore.vhd
--! @brief      Cipher core template module
--! @project    CAESAR Candidate Evaluation
--! @author     Ekawat (ice) Homsirikamol
--! @copyright  Copyright (c) 2016 Cryptographic Engineering Research Group
--!             ECE Department, George Mason University Fairfax, VA, U.S.A.
--!             All rights Reserved.
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
--! @note       This is publicly available encryption source code that falls
--!             under the License Exception TSU (Technology and software-
--!             —unrestricted)
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.ALL;
use ieee.std_logic_unsigned.all; 
use work.AEAD_pkg.all;
use ieee.numeric_std.all;

entity CipherCore is
    generic (
        --! Reset behavior
        G_ASYNC_RSTN    : boolean := False; --! Async active low reset
        --! Block size (bits)
        G_DBLK_SIZE     : integer := 128;   --! Data
        G_KEY_SIZE      : integer := 128;   --! Key
        G_TAG_SIZE      : integer := 128;   --! Tag
        --! The number of bits required to hold block size expressed in
        --! bytes = log2_ceil(G_DBLK_SIZE/8)
        G_LBS_BYTES     : integer := 4;
		  
		  --! Single pass authenticated ciphers max blocks: 27 (2^5 bytes per msg block = 2^32 bytes)
		  --! Two-pass authenticated ciphers max blocks: 6 (2^5 bytes per msg block = 2^11 bytes)
		  
		  --! For data limit of 2^32 - 1 set:
		  -- G_MAX_LEN = 32
		  -- G_MAX_L_INDEX = b"11010" or '1' & x"A" 
		  -- G_INDEX_WIDTH = 5
		  
		  --! For data limit of 2^11 - 1 set:
		  -- G_MAX_LEN = 11
		  -- G_MAX_L_INDEX = b"00101" or '0' & x"5" 
		  -- G_INDEX_WIDTH = 3
		  
		  G_MAX_LEN		: integer := 32;
		  --! Single pass authenticated ciphers max ntz index: 26
		  --! Two-pass authenticated ciphers max ntz index: 5
		  --! WARNING!  G_MAX_L_INDEX x 32 byte DRAMs or BRAMs will be instantiated
		  --! For example: 5 x 32 = 160 bytes of DRAM to hold precomputed L_ntz
		  G_MAX_L_INDEX   : std_logic_vector(4 downto 0) := '1' & x"A"; -- number of L[i] = L[0] .. L[{max} - 1]
		  G_INDEX_WIDTH   : integer := 5 		
    );
    port (
        --! Global
        clk             : in  std_logic;
        rst             : in  std_logic;
        --! PreProcessor (data)
        key             : in  std_logic_vector(G_KEY_SIZE       -1 downto 0);
        bdi             : in  std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        --! PreProcessor (controls)
        key_ready       : out std_logic;
        key_valid       : in  std_logic;
        key_update      : in  std_logic;
        decrypt         : in  std_logic;
        bdi_ready       : out std_logic;
        bdi_valid       : in  std_logic;
        bdi_type        : in  std_logic_vector(3                -1 downto 0);
        bdi_partial     : in  std_logic;
        bdi_eot         : in  std_logic;
        bdi_eoi         : in  std_logic;
        bdi_size        : in  std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        bdi_valid_bytes : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        bdi_pad_loc     : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        --! PostProcessor
        bdo             : out std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        bdo_valid       : out std_logic;
        bdo_ready       : in  std_logic;
        bdo_size        : out std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        msg_auth_done   : out std_logic;
        msg_auth_valid  : out std_logic
    );
end entity CipherCore;

architecture structure of CipherCore is

signal load_key, load_npub, sel_data : std_logic;
signal tag_valid : std_logic;

signal block_num_init, en_block_num : std_logic;
signal	H_sel		: std_logic;
signal	hash_sel	: std_logic_vector(1 downto 0);
signal	data_sel	: std_logic_vector(1 downto 0);
signal	delta_sel	: std_logic_vector(1 downto 0);
signal	tag_a_sel	: std_logic;
signal   compute_delta_n10 : std_logic;
signal	l_sel			: std_logic;
signal   output_sel  : std_logic;
signal   zero_sel		: std_logic;

signal	Fk_start	: std_logic;
signal	Fk_done		: std_logic;
	
signal	ld_lstar	: std_logic;
signal	l_start		: std_logic;
signal	l_load		: std_logic;
signal	l_done		: std_logic;
signal	ld_delta	: std_logic;
signal	ld_tag_a	: std_logic;
signal	ld_tag_e	: std_logic;

signal   no_AD		: std_logic;

signal last_block : std_logic;
signal block_full : std_logic;
signal en : std_logic;
signal decrypt_init, decrypt_reg : std_logic;
signal load_left, right_zeros : std_logic;

begin

OMD256_datapath: entity work.Datapath(structural)
		generic map(
		G_ASYNC_RSTN => G_ASYNC_RSTN,
		G_DBLK_SIZE => G_DBLK_SIZE,
		G_KEY_SIZE	=> G_KEY_SIZE,
		G_TAG_SIZE  => G_TAG_SIZE,
		G_LBS_BYTES => G_LBS_BYTES,
		G_MAX_LEN => G_MAX_LEN,
		G_MAX_L_INDEX => G_MAX_L_INDEX,
		G_INDEX_WIDTH => G_INDEX_WIDTH)
		port map(

	clk => clk,		
	rst => rst,	 
	en => en,   
	
-- data signals

	key => key,		
	bdi => bdi,		
	bdo => bdo,		
	tag_valid => tag_valid,		
	bdi_size	=> bdi_size,
	bdi_type => bdi_type,
	bdi_valid_bytes => bdi_valid_bytes,
	bdi_pad_loc => bdi_pad_loc,
	
-- control signals

	H_sel	=> H_sel,
	hash_sel => hash_sel,
	data_sel	=> data_sel,
	delta_sel => delta_sel,
	tag_a_sel => tag_a_sel,	
	output_sel => output_sel,
	zero_sel => zero_sel,
	compute_delta_n10 => compute_delta_n10,
	l_sel_in => l_sel,		
	decrypt => decrypt,	

	Fk_start	=> Fk_start,
	Fk_done => Fk_done,	
	
	no_AD => no_AD,
	ld_lstar	=> ld_lstar, 
	l_start => l_start,	
	l_load => l_load,		
	l_done => l_done,		
	ld_delta => ld_delta,	
	ld_tag_a => ld_tag_a,	
	ld_tag_e => ld_tag_e,	

	load_key => load_key,	
	load_npub => load_npub,	
	sel_data => sel_data,	
	last_block => last_block,	
	block_full_out => block_full,	
	decrypt_init => decrypt_init,
	decrypt_reg_out => decrypt_reg,
	load_left => load_left,	
	right_zeros => right_zeros,	
	block_num_init => block_num_init, 
	en_block_num => en_block_num 
);

OMD_ctlr: entity work.Controller(behavioral)
	port map(
		clk => clk,
		rst => rst,
		en => en,
		
	--! PreProcessor (controls)
      key_ready => key_ready,
		key_valid => key_valid,
		decrypt_init => decrypt_init,
		decrypt => decrypt_reg,
		key_update => key_update,
	   bdi_ready => bdi_ready,
		bdi_valid => bdi_valid,
		bdi_eot => bdi_eot,
		bdi_eoi => bdi_eoi,
		bdi_type => bdi_type,
	
   --! PostProcessor
      bdo_valid => bdo_valid,
		bdo_ready => bdo_ready,
      msg_auth_done => msg_auth_done,
		msg_auth_valid => msg_auth_valid,
	
   --! Datapath
	
		hash_sel => hash_sel,
		data_sel	=> data_sel,
		delta_sel => delta_sel,
		tag_a_sel => tag_a_sel,
		compute_delta_n10 => compute_delta_n10,
		l_sel => l_sel,
		H_sel => H_sel,
		output_sel => output_sel,
		zero_sel => zero_sel,

		Fk_start	=> Fk_start,
		Fk_done => Fk_done,	
	
		no_AD => no_AD,
		ld_delta => ld_delta,
		ld_tag_a => ld_tag_a,
		ld_tag_e => ld_tag_e,
		ld_lstar	=> ld_lstar, 
		l_start => l_start,	
		l_load => l_load,		
		l_done => l_done,
	
		block_num_init => block_num_init,
		en_block_num => en_block_num,
		last_block => last_block,
      tag_valid => tag_valid,
		load_key => load_key,
		load_npub => load_npub,
		load_left => load_left,
		right_zeros => right_zeros,
		sel_data => sel_data
	);

end structure;