
-------------------------------------------------------------------------------
--! @file       Controller.vhd
--! @author     William Diehl
--! @brief      OMD Controller
--! @date       8 July 2016
-------------------------------------------------------------------------------

LIBRARY ieee;
USE ieee.std_logic_1164.all;
USE ieee.std_logic_unsigned.all;

ENTITY Controller IS

PORT(

        clk: in std_logic;
		  rst: in std_logic;
	     en : out std_logic;

	--! PreProcessor (controls)
        key_ready       : out std_logic;
        key_valid       : in  std_logic;
        key_update      : in  std_logic;
        decrypt         : in  std_logic;
		  decrypt_init		: out std_logic;
        bdi_ready       : out std_logic;
        bdi_valid       : in  std_logic;
        bdi_type        : in  std_logic_vector(3                -1 downto 0);
        --bdi_partial     : in  std_logic;
        bdi_eot         : in  std_logic;
        bdi_eoi         : in  std_logic;
        --bdi_size        : in  std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        --bdi_valid_bytes : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        --bdi_pad_loc     : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);

        --! PostProcessor
        bdo_valid       : out std_logic;
        bdo_ready       : in  std_logic;
   	
		  block_num_init : out std_logic;
	     en_block_num : out std_logic;
       
        last_block : out std_logic;
   
	     no_AD : out std_logic;
	
		  load_key : out std_logic;
		  load_npub : out std_logic;
		  load_left : out std_logic;
		  right_zeros : out std_logic;
		  sel_data: out std_logic;
		  hash_sel: out std_logic_vector(1 downto 0);
		  data_sel: out std_logic_vector(1 downto 0);
		  delta_sel : out std_logic_vector(1 downto 0);
		  tag_a_sel : out std_logic;
		  output_sel : out std_logic;
		  zero_sel	  : out std_logic;
		  compute_delta_n10: out std_logic;
		  l_sel		 : out std_logic;
		  H_sel 	 : out std_logic;
		  Fk_start : out std_logic;
		  Fk_done  : in std_logic;
		  ld_lstar : out std_logic;
		  l_start  : out std_logic;
		  l_load	: out std_logic;
		  l_done   : in std_logic;
		  ld_delta : out std_logic;
		  ld_tag_a : out std_logic;
		  ld_tag_e : out std_logic;
		

	     tag_valid : IN STD_LOGIC;
	     msg_auth_valid : OUT STD_LOGIC;
	     msg_auth_done : OUT STD_LOGIC
	
	);
	
END Controller;

ARCHITECTURE behavioral OF Controller IS

SIGNAL PS, NS : STD_LOGIC_VECTOR(7 DOWNTO 0);
signal no_M, no_M_init, load_no_M : std_logic;
signal null_msg, null_msg_init, load_null_msg : std_logic;

BEGIN

sync_process: PROCESS(clk, rst)
BEGIN

-- Requires 1 asynchronous reset to put in State 0

IF (rst = '1') THEN
	PS <= x"00";
	
-- All state transitions occur on rising edge
-- All transitions are disabled if Read/Write Enable is low

ELSIF (rising_edge(clk)) 
	 THEN PS <= NS;
	     
END IF;
	
END Process;

public_process: PROCESS(PS, key_update, key_valid, bdi_valid, bdo_ready, bdi_eot,
                        decrypt, tag_valid, bdi_type, no_M, Fk_done, l_done)
BEGIN
 
	 -- defaults
block_num_init <= '0';
en_block_num <= '0';
last_block <= '0';

no_AD <= '0';
en <= '0';
decrypt_init <= '0';
ld_delta <= '0';
ld_tag_a <= '0';
ld_tag_e <= '0';
ld_lstar <= '0';
compute_delta_n10 <= '0';
l_sel <= '0';
output_sel <= '0';
zero_sel <= '0';

Fk_start <= '0';
l_start <= '0';
l_load <= '0';
no_M_init <= '0';
load_no_M <= '0';
load_npub <= '0';
load_key <= '0';
load_left <= '0';
right_zeros <= '0';
null_msg_init <= '0';
load_null_msg <= '0';
key_ready <= '0';
bdi_ready <= '0';
bdo_valid <= '0';
msg_auth_valid <= '0';
msg_auth_done <= '0';

case PS is
		 		 
	 WHEN x"00" => -- s_init
-- Initialization after asynchronous reset

		 	block_num_init <= '1';
         no_M_init <= '1';
			null_msg_init <= '1';
			delta_sel <= "00";
			tag_a_sel <= '0';
			ld_delta <= '1';
			ld_tag_a <= '1';
			H_sel <= '0'; -- latch
		   sel_data <= '1'; -- latch
			
			NS <= x"01";
				 
	 WHEN x"01" => -- s_key_update
	 -- Key check
		decrypt_init <= '1';
		if (bdi_valid = '1') then
			if (key_update = '1') then
				 if (key_valid = '1') then
						key_ready <= '1';
						load_key <= '1';
						NS <= x"02"; -- initialize L arrays = key initialization
				 else
					   NS <= x"01";
				 end if;
			else 
				 NS <= x"06"; 
			end if;
		else NS <= x"01";
		
		end if;
--!
--! compute lstar and  L arrays

	WHEN x"02" => -- s_init_l_array initialize L arrays
	-- must be called at least one cycle after load_key is asserted
	
		hash_sel <= "00"; -- align for lstar compute
		data_sel <= "01"; -- align for lstar compute
		Fk_start <= '1'; -- start lstar computation
		en <= '1';
	
		NS <= x"03"; 
	
	WHEN x"03" => -- s_compute_lstar
	
		if (Fk_done = '1') then
			ld_lstar <= '1';
			
			NS <= x"04"; 
		else
		   en <= '1';
			NS <= x"03"; 
		end if;
		
		
	WHEN x"04" => -- s_start_l_array
	-- must be called at least one cycle after ld_lstar is asserted
	
		l_start <= '1';
		l_load <= '1';
		NS <= x"05"; 
		
	WHEN x"05" => -- s_finish_l_array
	
		if (l_done = '1') then 
			
			NS <= x"06"; 
		else
			l_load <= '1';
			NS <= x"05"; 
		end if;
	
--! Finish L arrays

	WHEN x"06" => -- s_load_npub
	     	load_npub <= '1';
				if (bdi_eot = '1' and bdi_eoi = '1') then
					load_null_msg <= '1';
					no_AD <= '1'; 
					ld_delta <= '1'; 
					delta_sel <= "11"; 					
					NS <= x"07";  				
				else
					bdi_ready <= '1'; 
					NS <= x"08"; 
			   end if;

WHEN x"07" => -- s_tag_a_null_AD

		bdi_ready <= '1';
		no_AD <= '1'; 
		Fk_start <= '1'; -- tag_a must still be computed on null AD
		hash_sel <= "11";
		data_sel <= "10";
		en <= '1';
		NS <= x"0f"; 
	
-- associated data				
				
	WHEN x"08" => -- s_load_left

		if (bdi_valid = '1') then
			if (bdi_type(1) = '1') then -- process null AD
				
				no_AD <= '1';
				ld_delta <= '1'; 
				delta_sel <= "11";
				
				NS <= x"09"; 
			else
			
				load_left <= '1';	
				if (bdi_eot = '1') then
										
					load_no_M <= '1';
					last_block <= '1';
					delta_sel <= "11";
					ld_delta <= '1';
					right_zeros <= '1';
				
					NS <= x"0a";  
				else
					bdi_ready <= '1'; 
					NS<= x"0b"; 
				end if;
			end if;
		else 
			NS <= x"08"; 
		end if;
	
	WHEN x"09" => -- s_tag_a_null_AD2
	
		no_AD <= '1';
		Fk_start <= '1'; -- tag_a must still be computed on null AD
		en <= '1';
		hash_sel <= "11";
		data_sel <= "10";
		NS <= x"0f"; 

	WHEN x"0a" => -- s_compute_AD
	
		bdi_ready <= '1';
		last_block <= '1';
		right_zeros <= '1';
		hash_sel <= "11";
		data_sel <= "10";
		Fk_start <= '1';
		en <= '1';
		NS <= x"0e"; 		
	
   WHEN x"0b" => -- s_load_right
	
		if (bdi_valid = '1') then
		
			delta_sel <= "11";
			ld_delta <= '1';
			NS <= x"0c"; 
         if (bdi_eot = '1') then
				load_no_M <= '1';
				last_block <= '1';
			end if;	
		else
			NS <= x"0b"; 
		end if;
		
  WHEN x"0c" => -- s_start_right
  
		bdi_ready <= '1';
		hash_sel <= "11";
		data_sel <= "10";
		Fk_start <= '1';	
		en <= '1';
		if (bdi_eot = '1') then
			last_block <= '1';
			NS <= x"0e"; 
		else 
			NS <= x"0d"; 
		end if;
		
  WHEN x"0d" => -- s_finish_AD
  
		if (Fk_done = '1') then 
			ld_tag_a <= '1';
			tag_a_sel <= '1';
			en_block_num <= '1'; 
	      
			NS <= x"08"; 
		else
			en <= '1';
		   NS <= x"0d"; 
		end if;

  WHEN x"0e" => -- s_finish_last_AD
  
	  	if (Fk_done = '1') then
			ld_tag_a <= '1';
			tag_a_sel <= '1';
			
			Fk_start <= '1'; 
			hash_sel <= "01";
			data_sel <= "00";
			en <= '1';
										-- compute initial delta 
			NS <= x"10"; 
			
		else
			en <= '1';
		   last_block <= '1';
		   NS <= x"0e"; 
		end if;

	WHEN x"0f" => -- s_finish_null_AD
	
		if (Fk_done = '1') then
			ld_tag_a <= '1';
			tag_a_sel <= '1';
			Fk_start <= '1'; 
			hash_sel <= "01";
			data_sel <= "00";
			en <= '1';
										-- compute initial delta 
			NS <= x"10"; 
		else
			en <= '1';
			NS <= x"0f"; 
		end if;
		
	WHEN x"10" => -- s_init_delta_n00 initialize delta_n00
		if (Fk_done = '1') then
			delta_sel<= "10";
			block_num_init <= '1'; 
			ld_delta <= '1';
			NS <= x"11"; 
		else 
			en <= '1';
			NS <= x"10"; 
		end if;

	WHEN x"11" => -- s_load_delta_n10
	
		--! must be called at least one cycle after init delta_n00
		 
		 compute_delta_n10 <= '1';
		 delta_sel <= "11";
		 ld_delta <= '1';
		 NS <= x"12"; 
		 
	WHEN x"12" => -- s_start_delta_n10
	
		hash_sel <= "10";
		data_sel <= "01";
		H_sel <= '0';
		en_block_num <= '1';
		Fk_start <= '1';
		en <= '1';
		compute_delta_n10 <= '1';
		 
		if (no_M = '1' or null_msg = '1') then
			NS <= x"13"; 
		else
			NS <= x"15"; 
		end if;
	
	WHEN x"13" => -- s_load_null_M
	
		if (Fk_done = '1') then
			
			l_sel <= '1';
			zero_sel <= '1';
			delta_sel <= "11";
			ld_delta <= '1';
			
			NS <= x"14"; 
		
		else
			en <= '1';
			NS <= x"13"; 
		end if;
	
	WHEN x"14" => -- s_start_null_M
	
		l_sel <= '1';
		zero_sel <= '1';
		H_sel <= '1';
		hash_sel <= "10";
		data_sel <= "00";
		Fk_start <= '1';
		en <= '1';
		NS <= x"19"; 	
	
	WHEN x"15" => -- s_process_M
	
		 if (Fk_done = '1') then
			if (bdi_valid = '1') then
				if (bdo_ready = '1') then
					sel_data <= '1';
					bdo_valid <= '1'; -- Ci = H xor Mi
					if (bdi_eot = '1') then
						last_block <= '1';
						l_sel <= '1'; 
						delta_sel <= "11";
						ld_delta <= '1';
						NS <= x"16"; 
						if (decrypt = '1') then
							output_sel <= '1';
							
						end if;
					else 
						l_sel <= '0'; 
						ld_delta <= '1';
						delta_sel <= "11";
						en_block_num <= '1';
						if (decrypt = '1') then
							output_sel <= '1';
						end if;
						NS <= x"16"; 
					end if;
				else -- bdo_ready = '0'
					NS <= x"17"; -- wait for bdo ready and continue
				end if;
			else -- bdi_valid = '0'
				NS <= x"18"; -- wait for bdi valid and continue
			end if;
		else -- Fk_done = '0'
			en <= '1';
			NS <= x"15"; 
		end if;
	
	WHEN x"16" =>  -- s_start_M
	
		if (decrypt = '1') then
				output_sel <= '1';
							
		end if;
		
		if (bdi_eot = '1') then
				last_block <= '1'; 
				Fk_start <= '1'; 
				en <= '1'; 
				l_sel <= '1';  
				H_sel <= '1';  
				hash_sel <= "10"; 
				data_sel <= "11"; 
				bdi_ready <= '1'; 
				NS <= x"19"; -- process tag
		else -- bdi_eot = '0'
				bdi_ready <= '1'; 
				Fk_start <= '1'; 
				en <= '1'; 
				hash_sel <= "10";  
				data_sel <= "11";  
				H_sel <= '1';
				l_sel <= '0'; 
				NS <= x"15"; 
		end if;
	
	WHEN x"17" => -- s_bdo_ready wait for bdo ready and continue

		if (bdo_ready = '1') then
			bdo_valid <= '1';
			if (bdi_eot = '1') then
				last_block <= '1';
				l_sel <= '1';
				delta_sel <= "11";
				ld_delta <= '1';
				NS <= x"16"; 
				if (decrypt = '1') then
					output_sel <= '1';
					
				end if;
			else 
				l_sel <= '0';
				ld_delta <= '1';
				delta_sel <= "11";
				en_block_num <= '1';
				if (decrypt = '1') then
					output_sel <= '1';
				end if;
				NS <= x"16"; 
			end if;
		else -- bdo_valid = '0'
			NS <= x"17"; 
		end if;
		
		WHEN x"18" => -- s_bdi_valid wait for bdi valid and continue 
		
		if (bdi_valid = '1') then
			if (bdo_ready = '1') then
				sel_data <= '1';
				bdo_valid <= '1'; -- Ci = H xor Mi
				if (bdi_eot = '1') then
					last_block <= '1';
					l_sel <= '1';
					delta_sel <= "11";
					ld_delta <= '1'; 
					NS <= x"16"; 
					if (decrypt = '1') then
						output_sel <= '1';
						
					end if;
				else 
					l_sel <= '0';
					ld_delta <= '1';
					delta_sel <= "11";
					en_block_num <= '1';
					if (decrypt = '1') then
						output_sel <= '1';
					end if;
					NS <= x"16"; 
				end if;
			else -- bdo_ready = '0'
				NS <= x"17"; -- wait for bdo ready and continue
			end if;
		else -- bdi_valid = '0'
			NS <= x"18"; -- wait for bdi valid and continue
		end if;
			
	when x"19" => -- s_store_tag_e
	
		if (Fk_done = '1') then
			ld_tag_e <= '1';
			if (decrypt = '1') then		
				NS <= x"1a"; -- load and verify tag
			else
				NS <= x"1b"; -- generate tag
			end if;
		else
			en <= '1';
			NS <= x"19"; 
		end if;
		
	when x"1a" => -- s_verify_tag
		
		if (bdi_valid = '1') then
			msg_auth_done <= '1';
			if (tag_valid = '1') then
				msg_auth_valid <= '1';
			end if;
			bdi_ready <= '1';
			NS <= x"00";
		else
			NS <= x"1a"; 
		end if;
	
   when x"1b" => -- s_generate_tag

		if (bdo_ready = '1') then
			sel_data <= '0';
			bdo_valid <= '1';
			NS <= x"00";
		else
			NS <= x"1b"; 
		end if;	
					
	WHEN OTHERS =>
	
		  NS <= x"00";
			  
	  END case; 
END Process;

store_no_M: entity work.reg1(behavioral)
	port map(
		d => bdi_eoi,
		reset => rst,
		clk => clk,
		init => no_M_init,
		en => load_no_M,
		q => no_M);

store_null_msg: entity work.reg1(behavioral)
	port map(
		d => bdi_eoi,
		reset => rst,
		clk => clk,
		init => null_msg_init,
		en => load_null_msg,
		q => null_msg);
		
END behavioral;  