
-------------------------------------------------------------------------------
--! @file       Datapath.vhd
--! @author     William Diehl
--! @brief      OMD Datapath
--! @date       8 July 2016
-------------------------------------------------------------------------------


-- OMD Datapath

library ieee;
use ieee.std_logic_1164.all; 
use ieee.std_logic_unsigned.all;  
use IEEE.NUMERIC_STD.ALL;

entity Datapath is

generic (
        --! Reset behavior
        G_ASYNC_RSTN    : boolean := False; --! Async active low reset
        --! Block size (bits)
        G_DBLK_SIZE     : integer := 128;   --! Data
        G_KEY_SIZE      : integer := 128;   --! Key
        G_TAG_SIZE      : integer := 128;   --! Tag
        --! The number of bits required to hold block size expressed in
        --! bytes = log2_ceil(G_DBLK_SIZE/8)
        G_LBS_BYTES     : integer := 4;
		  G_MAX_LEN		: integer := 32;
		  G_MAX_L_INDEX   : std_logic_vector(4 downto 0) := '1' & x"A"; -- maximum index i for L[i]
		  G_INDEX_WIDTH	: integer := 5
    );

port(

	clk		:in std_logic;
	rst		:in std_logic; 
	en       :in std_logic;

-- data signals

	key		:in std_logic_vector(G_KEY_SIZE - 1 downto 0);
	bdi		:in std_logic_vector(G_DBLK_SIZE - 1 downto 0);
	bdo		:out std_logic_vector(G_DBLK_SIZE - 1 downto 0);
	tag_valid	:out std_logic;	
	bdi_size	:in std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
	bdi_type :in std_logic_vector(2 downto 0);
	bdi_valid_bytes : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
   bdi_pad_loc     : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);

-- control signals

	H_sel			:in std_logic;
	hash_sel		:in std_logic_vector(1 downto 0);
	data_sel		:in std_logic_vector(1 downto 0);
	delta_sel	:in std_logic_vector(1 downto 0);
	tag_a_sel	:in std_logic;
	output_sel  :in std_logic;
	zero_sel		:in std_logic;
	l_sel_in		:in std_logic;
	compute_delta_n10: in std_logic;
	decrypt		:in std_logic;

	Fk_start		:in std_logic;
	Fk_done		:out std_logic;
	
	ld_lstar		:in std_logic;
	l_start		:in std_logic;
	l_load		:in std_logic;
	l_done		:out std_logic;
	ld_delta		:in std_logic;
	ld_tag_a		:in std_logic;
	ld_tag_e		:in std_logic;

	no_AD		   :in std_logic;
	load_key		:in std_logic;
	load_npub	:in std_logic;
	sel_data		:in std_logic;
	last_block	:in std_logic;
	block_full_out	:out std_logic;
	decrypt_init:in std_logic;
	decrypt_reg_out:out std_logic;
	load_left	:in std_logic;
	right_zeros	:in std_logic;
	block_num_init :in std_logic;
	en_block_num   :in std_logic
   
);

end Datapath;

architecture structural of Datapath is
 
constant G_MAX_BLOCK : integer:= 2**(G_MAX_LEN-5); -- maximumn number of blocks
constant G_MAX_BLOCK_SIZE : integer:=(G_MAX_LEN-5);  --ceil[log2(G_MAX_BLOCK)]

constant zeros256: std_logic_vector(255 downto 0):= (OTHERS => '0'); 
constant zeros128: std_logic_vector(127 downto 0):= (OTHERS => '0'); 

constant tau256: std_logic_vector(255 downto 0):= x"0000000000000000000000000000000000000000000000000000000000000080"; -- |tau|=128 bits
constant zeros31 : std_logic_vector(251 downto 0):= (OTHERS => '0'); 

signal npub256, zeros: std_logic_vector(255 downto 0);
signal H, delta_xor_H, delta_xor_left, left, right, msg, msg_data: std_logic_vector(255 downto 0);
signal hash, res: std_logic_vector(255 downto 0);
signal data: std_logic_vector(511 downto 0);
signal lstar, l_ntz, lstar_2, lstar_3, l_in, l_last: std_logic_vector(255 downto 0);
signal next_delta, delta: std_logic_vector(255 downto 0);
signal delta_xor_l_in : std_logic_vector(255 downto 0); 
signal res_xor_tag_a, next_tag_a, tag_a, tag_e: std_logic_vector(G_TAG_SIZE - 1 downto 0); 
signal tag_a_xor_tag_e: std_logic_vector(G_TAG_SIZE - 1 downto 0);
signal left_reg : std_logic_vector(255 downto 0);

signal decrypt_reg : std_logic;
signal block_full : std_logic;
signal index: std_logic_vector(G_INDEX_WIDTH - 1 downto 0);

signal right_sel : std_logic_vector(1 downto 0);

signal l_sel : std_logic_vector(1 downto 0);
signal ld_tag_a_delay: std_logic;
signal Fk_pre_done : std_logic; -- goes high one cycle before Fk_done
-- output from sha256_comp64 "res" is valid at Fk_done

signal K : std_logic_vector(G_KEY_SIZE - 1 downto 0);
signal npub : std_logic_vector(95 downto 0);
signal right_zero_input : std_logic_vector(G_DBLK_SIZE - 1 downto 0);

signal tag, exp_tag : std_logic_vector(G_TAG_SIZE - 1 downto 0);

signal next_block_number, block_number : std_logic_vector(G_MAX_BLOCK_SIZE - 1 downto 0);

signal output, output_decrypt, output_padded, 
		 output_masking, output_padding: std_logic_vector(G_DBLK_SIZE - 1 downto 0);
 
begin

-- store key and npub during initialization

store_key: entity work.regn_en(behavioral)
	generic map(n=>G_KEY_SIZE)
	port map(
		d => key,
		reset => rst,
		clk => clk,
		init => '0',
		en => load_key,
		q => K);

store_npub: entity work.regn_en(behavioral)
	generic map(n=>96)
	port map(
		d => bdi(G_DBLK_SIZE - 1 downto G_DBLK_SIZE - 96),
		reset => rst,
		clk => clk,
		init => '0',
		en => load_npub,
		q => npub);

store_left: entity work.regn_en(behavioral)
	generic map(n=>G_DBLK_SIZE)
	port map(
		d => bdi,
		reset => rst,
		clk => clk,
		init => '0',
		en => load_left,
		q => left_reg);

store_block_number: entity work.regn_en(behavioral)
	generic map(n=>G_MAX_BLOCK_SIZE)
	port map(
		d => next_block_number,
		reset => rst,
		clk => clk,
		init => block_num_init,
		en => en_block_num,
		q => block_number);

store_decrypt: entity work.reg1(behavioral)
	port map(
		d => decrypt,
		reset => rst,
		clk => clk,
		init => '0',
		en => decrypt_init,
		q => decrypt_reg);
		
decrypt_reg_out <= decrypt_reg;

-- the Fk_pre_done signal would enable pipelined operations that require knowledge
-- of the "next to last" round"
	
store_Fk_pre_done: entity work.reg1(behavioral)
	port map(
		d => Fk_pre_done,
		reset => rst,
		clk => clk,
		init => '0',
		en => '1',
		q => Fk_done);

	next_block_number <= block_number + 1;

	block_full <= bdi_size(G_LBS_BYTES);
	block_full_out <= block_full;
		
	right_zero_input <= block_full & "000" & zeros31;
		
	right_sel(0) <= right_zeros;
	right_sel(1) <= no_AD; -- high when no AD segment
	
	with right_sel select
    		right <= bdi when "00",			
		           right_zero_input when "01",			
			        (x"0" & zeros31) when "10",			
	              (x"0" & zeros31) when "11", 
	              bdi when others;

	left <= left_reg when (no_AD = '0') else (x"8" & zeros31);
	
   npub256 <= npub & x"8000000000000000000000000000000000000000"; -- N || 1||0^159

   H <= zeros256 when (H_sel = '0') else res;
	delta_xor_H <= delta xor H;
	delta_xor_left <= delta xor left;

	with hash_sel select
    		hash <= zeros256 when "00",			
		           npub256 when "01",			
			        delta_xor_H when "10",			
	              delta_xor_left when "11", -- AD
	              zeros256 when others;

	msg <= bdi when (output_sel = '0') else output_decrypt;
	
	output_decrypt <= output when (block_full = '1') else output_padded;
	output_padded <= (output and output_masking) xor (output_padding);
	
	g1: for i in 0 to 31 generate
		output_masking((31 - i) * 8 + 7 downto (31 - i) * 8) <= bdi_valid_bytes(31-i) & 
		bdi_valid_bytes(31-i) & bdi_valid_bytes(31-i) & bdi_valid_bytes(31-i) & bdi_valid_bytes(31-i) &
		bdi_valid_bytes(31-i) & bdi_valid_bytes(31-i) & bdi_valid_bytes(31-i);
	end generate g1;
	
	g2: for i in 0 to 31 generate
		output_padding((31 - i) * 8 + 7 downto (31 - i) * 8) <= bdi_pad_loc(31-i) & "0000000";
	end generate g2;
	
	zeros <= zero_sel & "000" & zeros31;
	
	with data_sel select
    		msg_data <= zeros when "00",			
		               tau256 when "01",			
			            right when "10", -- AD			
	                  msg when "11",
	                  zeros when others;

        data <= K & x"00000000000000000000000000000000" & msg_data; -- key || 0^128 || msg

	Fk: entity work.sha256_comp64(structural)
        port map (
    		start => Fk_start,
    		done => Fk_pre_done,
    		clk => clk, 
    		rst => rst,
			en => en,
			hash => hash,
			data => data,
			res => res);

	store_lstar: entity work.regn_en(behavioral)
	generic map(n=>256)
	port map(
		d => res,
		reset => rst,
		clk => clk,
		init => '0',
		en => ld_lstar,
		q => lstar);

-- compute ntz[block_number]
-- ntz = "number of trailing zeros"
	
	get_ntz: entity work.ntz(functional)
		generic map (G_MAX_LEN => G_MAX_LEN,
					    G_INDEX_WIDTH => G_INDEX_WIDTH)
     	port map (
    		address => next_block_number,
    		output => index);

-- generate L_array using lstar as a seed
-- l_ntz = L[ntz(block_number)]

	store_L_array: entity work.L_array(structural)
        generic map(G_MAX_L_INDEX => G_MAX_L_INDEX,
						  G_INDEX_WIDTH => G_INDEX_WIDTH)
		  port map (
    		start => l_start,
    		done => l_done,
    		clk => clk, 
    		rst => rst,
			load => l_load,
			lstar => lstar,
			l_index => l_ntz,
			index => index);

	lstar_x2: entity work.GF2(dataflow)
     	port map (
    	x => lstar, 
    	o => lstar_2);

	lstar_3 <= lstar_2 xor lstar; 
	l_sel(1) <= l_sel_in; -- 1 = Ek or Dk; 0 = AD
	l_sel(0) <= (block_full and (not right_zeros)) or compute_delta_n10;
	-- by definition, if right_zeros = 1 then last block cannot be full

	with l_sel select
    		l_last <= lstar when "00",			
					    l_ntz when "01",			
	                lstar_3 when "10",			
	                lstar_2 when "11",
	                lstar when others;

	l_in <= l_last when (last_block = '1' or no_AD = '1' or zero_sel = '1') else l_ntz;

	delta_xor_l_in <= delta xor l_in; 
	
	with delta_sel select
    		next_delta <= zeros256 when "00",			
							  zeros256 when "01",			
							  res when "10",			
							  delta_xor_l_in when "11",
							  zeros256 when others;
	
	store_delta: entity work.regn_en(behavioral)
	generic map(n=>256)
	port map(
		d => next_delta,
		reset => rst,
		clk => clk,
		init => '0',
		en => ld_delta,
		q => delta 
		);
	
	res_xor_tag_a <= res(G_DBLK_SIZE - 1 downto G_TAG_SIZE)  xor tag_a;
		
	next_tag_a <= zeros128 when (tag_a_sel = '0') else res_xor_tag_a;
	
	store_tag_a: entity work.regn_en(behavioral)
	generic map(n=>128) 
	port map(
		d => next_tag_a,
		reset => rst,
		clk => clk,
		init => '0',
		en => ld_tag_a, 
		q => tag_a);

	store_tag_e: entity work.regn_en(behavioral)
	generic map(n=>128)  
	port map(
		d => res(G_DBLK_SIZE - 1 downto G_TAG_SIZE), 
		reset => rst,
		clk => clk,
		init => '0',
		en => ld_tag_e,
		q => tag_e);

	tag_a_xor_tag_e <= tag_a xor tag_e;
	
	tag <= tag_a_xor_tag_e;
	exp_tag <= bdi(G_DBLK_SIZE - 1 downto G_DBLK_SIZE - G_TAG_SIZE);
	tag_valid <= '1' when (exp_tag = tag_a_xor_tag_e) else '0';
	output <= bdi xor res; -- Ci = Mi xor Fk (Ek), Mi = Ci xor Fk (Dk)

	bdo <= output when (sel_data = '1') else (tag & zeros128);

end structural; 