
-------------------------------------------------------------------------------
--! @file       sha256_comp64.vhd
--! @author     William Diehl
--! @brief      OMD sha256_comp64
--! @date       8 July 2016
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all; 
use ieee.std_logic_unsigned.all;  

entity sha256_comp64 is
port(

	clk		:in std_logic;
	rst		:in std_logic; 
	en              :in std_logic;
	start		:in std_logic;

	hash		:in std_logic_vector(255 downto 0);
	data		:in std_logic_vector(511 downto 0);  
	
	done		:out std_logic;
	res		:out std_logic_vector(255 downto 0)
);

end sha256_comp64;

architecture structural of sha256_comp64 is

constant k_0 : std_logic_vector(31 downto 0):= x"428a2f98"; -- k 0 constant

signal  a_in,b_in,c_in,d_in,e_in,f_in,g_in,h_in: std_logic_vector(31 downto 0);
signal  a,b,c,d,e,f,g,h: std_logic_vector(31 downto 0);
signal  a0,b0,c0,d0,e0,f0,g0,h0: std_logic_vector(31 downto 0);

signal  next_a,next_b,next_c,next_d,next_e,next_f,next_g,next_h: std_logic_vector(31 downto 0);
signal  a_reg,b_reg,c_reg,d_reg,e_reg,f_reg,g_reg,h_reg: std_logic_vector(31 downto 0);

signal  next_x, next_x0, next_x1, next_x9, next_x14, x_in: std_logic_vector(31 downto 0);  

signal  x0_reg, x1_reg, x2_reg, x3_reg, x4_reg, x5_reg, x6_reg, x7_reg, x8_reg, x9_reg, x10_reg,
        x11_reg, x12_reg, x13_reg, x14_reg, x15_reg : std_logic_vector(31 downto 0);

signal  kHX_init, next_kHX, kHX_reg, kHX : std_logic_vector(31 downto 0);

signal  piso, next_piso: std_logic_vector(479 downto 0);

signal  k: std_logic_vector(31 downto 0);

signal  round_counter, next_round_counter: std_logic_vector(5 downto 0);
signal  sel_func: std_logic;

signal  X, s0, s1: std_logic_vector(31 downto 0);

begin
	a_in <= hash(255 downto 224); 
	b_in <= hash(223 downto 192);
	c_in <= hash(191 downto 160);
	d_in <= hash(159 downto 128);
	e_in <= hash(127 downto 96);
	f_in <= hash(95 downto 64);
	g_in <= hash(63 downto 32);
	h_in <= hash(31 downto 0);

	sel_func <= next_round_counter(5) or next_round_counter(4); -- sel_func = 1 if we are in round 16 or higher, else sel_func = 0
	next_round_counter <= round_counter + 1;
	done <= '1' when (round_counter = "111111") else '0';

-- initialize registers

	a <= a0 when (round_counter = 0) else a_reg;
	b <= b0 when (round_counter = 0) else b_reg;
	c <= c0 when (round_counter = 0) else c_reg;
	d <= d0 when (round_counter = 0) else d_reg;
	e <= e0 when (round_counter = 0) else e_reg;
	f <= f0 when (round_counter = 0) else f_reg;
	g <= g0 when (round_counter = 0) else g_reg;
	h <= h0 when (round_counter = 0) else h_reg;

-- compute round 0
-- compute K and X constants

	get_constant: entity work.sha2_constants(dataflow) 
	generic map(n=>32, a=>6)	
	port map(
		address=> next_round_counter,
  		output=> k);

-- compute Xi = Xi + s0(Xi+1) + xi+9 + s1(xi+14)

	next_x <= x_in when (sel_func = '0') else x15_reg;

	get_s0: entity work.sigma0_lower(dataflow) 
	generic map(n=>32)	
	port map(
		x=> next_x,
  		o=> s0);

	next_x1 <= x0_reg + s0;
	
	get_s1: entity work.sigma1_lower(dataflow) 
	generic map(n=>32)	
	port map(
		x=> next_x,
  		o=> s1);

	next_x9 <= x8_reg + next_x;
	
	next_x14 <= x13_reg + s1;
	
	X <= next_x;

-- update status registers
	
	update_regs: entity work.sha256_round(structural) 
	generic map(n=>32)	
	port map(
		a => a,
		b => b,
		c => c,
		d => d,
		e => e,
		f => f,
		g => g,
		h => h,
		X => X,
		K => k,
		kHX_in => kHX_reg,
		kHX_out => kHX,
		a_out => next_a,
		b_out => next_b,
		c_out => next_c,
		d_out => next_d,
		e_out => next_e,
		f_out => next_f,
		g_out => next_g,
		h_out => next_h);

-- align outputs

	res <= (a_reg + a0) & (b_reg + b0) & (c_reg + c0) & (d_reg + d0) &
 	       (e_reg + e0) & (f_reg + f0) & (g_reg + g0) & (h_reg + h0);

-- store results

	store_a: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => next_a,
		reset => rst,
		clk => clk,
		init => '0',
		en => en,
		q => a_reg);

	store_b: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => next_b,
		reset => rst,
		clk => clk,
		init => '0',
		en => en,
		q => b_reg);
	
	store_c: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => next_c,
		reset => rst,
		clk => clk,
		init => '0',
		en => en,
		q => c_reg);
	
	store_d: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => next_d,
		reset => rst,
		clk => clk,
		init => '0',
		en => en,
		q => d_reg);
		
	store_e: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => next_e,
		reset => rst,
		clk => clk,
		init => '0',
		en => en,
		q => e_reg);
	
	store_f: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => next_f,
		reset => rst,
		clk => clk,
		init => '0',
		en => en,
		q => f_reg);
	
	store_g: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => next_g,
		reset => rst,
		clk => clk,
		init => '0',
		en => en,
		q => g_reg);
	
	store_h: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => next_h,
		reset => rst,
		clk => clk,
		init => '0',
		en => en,
		q => h_reg);

-- store initial status registers
-- must sum initial status registers with results after 64 rounds

	store_a0: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => a_in,
		reset => rst,
		clk => clk,
		init => '0',
		en => start,
		q => a0);

	store_b0: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => b_in,
		reset => rst,
		clk => clk,
		init => '0',
		en => start,
		q => b0);
	
	store_c0: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => c_in,
		reset => rst,
		clk => clk,
		init => '0',
		en => start,
		q => c0);
	
	store_d0: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => d_in,
		reset => rst,
		clk => clk,
		init => '0',
		en => start,
		q => d0);
		
	store_e0: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => e_in,
		reset => rst,
		clk => clk,
		init => '0',
		en => start,
		q => e0);
	
	store_f0: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => f_in,
		reset => rst,
		clk => clk,
		init => '0',
		en => start,
		q => f0);
	
	store_g0: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => g_in,
		reset => rst,
		clk => clk,
		init => '0',
		en => start,
		q => g0);
	
	store_h0: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => h_in,
		reset => rst,
		clk => clk,
		init => '0',
		en => start,
		q => h0);

-- message scheduler

-- update data registers

next_x0 <= data(511 downto 480) when (start = '1') else next_x;

	store_x0: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => next_x0,
		reset => rst,
		clk => clk,
		init => '0',
		en => en,
		q => x0_reg);

	store_x1: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => next_x1,
		reset => rst,
		clk => clk,
		init => '0',
		en => en,
		q => x1_reg);
	
	store_x2: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => x1_reg,
		reset => rst,
		clk => clk,
		init => '0',
		en => en,
		q => x2_reg);
	
	store_x3: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => x2_reg,
		reset => rst,
		clk => clk,
		init => '0',
		en => en,
		q => x3_reg);
		
	store_x4: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => x3_reg,
		reset => rst,
		clk => clk,
		init => '0',
		en => en,
		q => x4_reg);
	
	store_x5: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => x4_reg,
		reset => rst,
		clk => clk,
		init => '0',
		en => en,
		q => x5_reg);
	
	store_x6: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => x5_reg,
		reset => rst,
		clk => clk,
		init => '0',
		en => en,
		q => x6_reg);
	
	store_x7: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => x6_reg,
		reset => rst,
		clk => clk,
		init => '0',
		en => en,
		q => x7_reg);

	store_x8: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => x7_reg,
		reset => rst,
		clk => clk,
		init => '0',
		en => en,
		q => x8_reg);
	
	store_x9: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => next_x9,
		reset => rst,
		clk => clk,
		init => '0',
		en => en,
		q => x9_reg);
	
	store_x10: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => x9_reg,
		reset => rst,
		clk => clk,
		init => '0',
		en => en,
		q => x10_reg);
	
	store_x11: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => x10_reg,
		reset => rst,
		clk => clk,
		init => '0',
		en => en,
		q => x11_reg);
		
	store_x12: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => x11_reg,
		reset => rst,
		clk => clk,
		init => '0',
		en => en,
		q => x12_reg);
	
	store_x13: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => x12_reg,
		reset => rst,
		clk => clk,
		init => '0',
		en => en,
		q => x13_reg);
	
	store_x14: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => next_x14,
		reset => rst,
		clk => clk,
		init => '0',
		en => en,
		q => x14_reg);
	
	store_x15: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => x14_reg,
		reset => rst,
		clk => clk,
		init => '0',
		en => en,
		q => x15_reg);

-- Precompute kHX

	kHX_init <= data(511 downto 480) + k_0 + h_in;

	next_kHX <= kHX_init when (start = '1') else kHX;

	store_kHX: entity work.regn_en(behavioral)
	generic map(n=>32)
	port map(
		d => next_KHX,
		reset => rst,
		clk => clk,
		init => '0',
		en => en,
		q => kHX_reg);

-- PISO for 512-bit data

	next_piso <= data(479 downto 0) when (start = '1') else (piso(447 downto 0) & x"00000000"); 
	x_in <= piso(479 downto 448);
	
	store_piso: entity work.regn_en(behavioral)
	generic map(n=>480)
	port map(
		d => next_piso,
		reset => rst,
		clk => clk,
		init => '0',
		en => en,
		q => piso);

-- update counter

	update_cnt: entity work.regn_en(behavioral)
	generic map(n=>6)
	port map(
		d => next_round_counter,
		reset => rst,
		clk => clk,
		init => start,
		en => en,
		q => round_counter);

end structural; 